package mirage-net-xen

  1. Overview
  2. Docs

mirage-net-xen

API

Library mirage-net-xen

  • Assemble Assemble complete network packets from Xen network messages.
  • Extra Descriptor types, from io/netif.h.
  • Features
  • Flags
  • Netif
  • RX
  • S
  • Shared_page_pool Pool and reuse shared memory blocks. This saves having to reallocate and reshare pages of memory, which is slow. A block is half the size of a page, since network packets usually don't require more than this.
  • Stats
  • TX
  • Xenstore